from aryth.bound_entries import max_by
from texting import SP, lange
from veho.entries import duozipper, trizipper

from pyspare.padder.pad_string import to_lpad, to_pad


def pad_entries(text, raw=None, dye=None, ansi=False, fill=SP):
    raw = raw if raw is not None else text
    lpad, pad = to_lpad(ansi=ansi, fill=fill), to_pad(ansi=ansi, fill=fill)
    kw, vw = max_by(text, lange if ansi else len)
    return trizipper(
        text, raw, dye,
        lambda tx, va, dy: dy(lpad(tx, kw)),
        lambda tx, va, dy: dy(pad(tx, vw, va))
    ) \
        if dye \
        else duozipper(
        text, raw,
        lambda tx, va: lpad(tx, kw),
        lambda tx, va: pad(tx, vw, va)
    )



from dataclasses import dataclass
from typing import List

from veho.matrix import margin_mapper, margin_mutate, margin_shallow, size
from veho.vector import init

from pyspare.margin.matrix_margin.sizing import sizing
from pyspare.margin.utils import marginal


@dataclass
class MatrixMargin:
    matrix: List[list]
    top: int
    bottom: int
    left: int
    right: int
    height: int
    width: int

    @staticmethod
    def build(mx, tp, bt, lf, rt, ht=None, wd=None) -> 'MatrixMargin':
        (tp, bt, lf, rt, ht, wd) = sizing(mx, tp, bt, lf, rt, ht, wd)
        cut_mx = margin_shallow(mx, tp, bt, lf, rt)
        return MatrixMargin(cut_mx, tp, bt, lf, rt, ht, wd)

    def empty_row(self, el) -> list:
        return init(self.left + self.right, lambda _: el)

    def map(self, fn, mutate=False) -> 'MatrixMargin':
        boot, mapper = (self.reboot, margin_mutate) if mutate else (self.clone, margin_mapper)
        mx, tp, bt, lf, rt = self.matrix, self.top, self.bottom, self.left, self.right
        return boot(mapper(mx, fn, tp, bt, lf, rt))

    def stringify(self, func=None, mutate=False) -> 'MatrixMargin':
        fn = (lambda x: str(func(x))) if func else str
        return self.map(fn, mutate)

    def to_matrix(self, el):
        mx, tp, bt, lf, rt = self.matrix, self.top, self.bottom, self.left, self.right
        dash_x, dash_y = bool(bt), bool(rt)
        upper = [marginal(row, lf, rt, dash_y and el) for row in mx[:tp]] if tp else []
        lower = [marginal(row, lf, rt, dash_y and el) for row in mx[-bt:]] if bt else []
        w = size(upper)[1] or size(lower)[1]
        blank_line = [[el] * w] if w and dash_x and el else []
        return upper + blank_line + lower

    def reboot(self, mx) -> 'MatrixMargin':
        if mx: self.matrix = mx
        return self

    def clone(self, mx) -> 'MatrixMargin':
        tp, bt, lf, rt, ht, wd = self.top, self.bottom, self.left, self.right, self.height, self.width
        return MatrixMargin(mx, tp, bt, lf, rt, ht, wd)



from typing import Callable, List, Tuple

from palett import Preset
from palett.presets import FRESH, PLANET
from texting import COLF, RTSP
from texting.enum.brackets import BRC, NONE

from pyspare.deco.deco_entries.deco_entries import deco_entries


def deco_dict(
        lex: dict,
        key_read: Callable = None,
        read: Callable = None,
        head: int = None,
        tail: int = None,
        presets: Tuple[Preset] = (FRESH, PLANET),
        effects: List[str] = None,
        delim: str = COLF,
        bracket: int = BRC,
        ansi: bool = False,
        dash: str = RTSP
):
    if not isinstance(lex, dict) or not dict: return str(lex)
    return deco_entries(list(lex.items()),
                        key_read=key_read,
                        read=read,
                        head=head,
                        tail=tail,
                        presets=presets,
                        effects=effects,
                        delim=delim,
                        bracket=bracket,
                        inner_bracket=NONE,
                        ansi=ansi,
                        dash=dash)



from typing import Callable, List, Tuple

from palett import Preset
from palett.presets import FRESH, PLANET
from texting import COLF, RTSP
from texting.enum.brackets import BRC

from pyspare.deco.deco_entries.deco_dict import deco_dict


def deco_object(
        ob: dict,
        key_read: Callable = None,
        read: Callable = None,
        presets: Tuple[Preset] = (FRESH, PLANET),
        effects: List[str] = None,
        delim: str = COLF,
        bracket: int = BRC,
        ansi: bool = False,
        dash: str = RTSP
):
    lex = ob.__dict__ \
        if hasattr(ob, '__dict__') \
        else {s: getattr(ob, s) for s in ob.__slots__} \
        if hasattr(ob, '__slots__') \
        else {}
    stringed = deco_dict(lex,
                         key_read=key_read,
                         read=read,
                         presets=presets,
                         effects=effects,
                         delim=delim,
                         bracket=bracket,
                         ansi=ansi,
                         dash=dash) if lex else str(ob)
    return f'[{type(ob).__name__}] {stringed}'



from veho.matrix import size


def sizing(mx, tp, bt, lf, rt, ht=None, wd=None):
    if not ht or not wd: (ht, wd) = size(mx)
    if not ht or not wd: tp, bt = 0, 0
    if (not tp and not bt) or (tp + bt >= ht): tp, bt = ht, 0
    if (not lf and not rt) or (lf + rt >= wd): lf, rt = wd, 0
    return tp, bt, lf, rt, ht, wd



from intype import is_numeric
from texting import has_ansi, lange
from texting.pad import lpad, rpad


def ansi_pad_len(tx, pd):
    return len(tx) + pd - lange(tx) if has_ansi(tx) else pd


def to_pad(fill, ansi=True, dock=False):
    if dock:
        padder = lpad if dock < 0 else rpad
        return (lambda tx, pd: padder(tx, ansi_pad_len(tx, pd), fill)) \
            if ansi \
            else (lambda tx, pd: padder(tx, pd, fill))
    else:
        return (lambda tx, pd, v: (lpad if is_numeric(v) else rpad)(tx, ansi_pad_len(tx, pd), fill)) \
            if ansi \
            else (lambda tx, pd, v: (lpad if is_numeric(v) else rpad)(tx, pd, fill))


def to_lpad(fill, ansi=True):
    return (lambda tx, pd: lpad(tx, ansi_pad_len(tx, pd), fill)) \
        if ansi \
        else (lambda tx, pd: lpad(tx, pd, fill))


def to_rpad(fill, ansi=True):
    return (lambda tx, pd: rpad(tx, ansi_pad_len(tx, pd), fill)) \
        if ansi \
        else (lambda tx, pd: rpad(tx, pd, fill))



from texting import lange


def mutate_key_pad(entries):
    wd, hi = 0, len(entries)
    while (hi := hi - 1) >= 0:
        k, v = entries[hi]
        wd = max(lange(k := str(k)), wd)
        entries[hi] = k, v
    return wd



def marginal(ve, h, t, ellip):
    return (((ve[:h] + [ellip] + ve[-t:]) if ellip else (ve[:h] + ve[-t:]))
            if t
            else ve[:h]) if h else ((([ellip] + ve[-t:]) if ellip else ve[-t:])
                                    if t
                                    else [])



from dataclasses import dataclass
from typing import Tuple

from palett import Preset
from palett.presets import AZURE, MOSS


@dataclass
class DecoPreset:
    presets: Tuple[Preset, Preset] or Preset = (AZURE, MOSS)
    depth: int = 8
    vert: int = 0
    unit: int = 32
    width: int = 80



from aryth.bound_vector import max_by
from texting import SP, lange
from veho.columns import mapper
from veho.matrix.enumerate import duozipper, trizipper

from pyspare.padder.pad_string import to_pad


def pad_matrix(text, raw=None, dye=None, ansi=False, fill=SP):
    raw = raw if raw is not None else text
    pad = to_pad(ansi=ansi, fill=fill)
    length = lange if ansi else len
    wds = mapper(text, lambda col: max_by(col, indicator=length))
    return trizipper(text, raw, dye, lambda tx, va, dy, i, j: dy(pad(tx, wds[j], va))) \
        if dye \
        else duozipper(text, raw, lambda tx, va, i, j: pad(tx, wds[j], va))



from aryth.bound_vector import max_by
from texting import SP, lange
from veho.vector import duozipper, trizipper

from pyspare.padder.pad_string import to_pad


def pad_vector(text, raw=None, dye=None, ansi=False, fill=SP):
    raw = raw if raw is not None else text
    pad = to_pad(ansi=ansi, fill=fill)
    wd = max_by(text, lange if ansi else len)
    return trizipper(text, raw, dye, lambda tx, va, dy: dy(pad(tx, wd, va))) \
        if dye \
        else duozipper(text, raw, lambda tx, va: pad(tx, wd, va))



def pad_start(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}>{width}}'


def pad_end(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}<{width}}'


def pad_centered(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}^{width}}'



import inspect

from intype import is_numeric
from palett import fluo_entries, fluo_vector
from texting import brace, bracket, parenth
from veho.entries import mutate_values
from veho.vector import mutate

from pyspare.deco.deco_node.helpers.mutate_key_pad import mutate_key_pad
from pyspare.deco.deco_node.preset import DecoPreset
from pyspare.deco.deco_node.render import render_entries, render_vector


def deco(ob): return deco_node(DecoPreset(), ob, 0)


def deco_node(self, node, level=0, indent=0):
    if node is None: return str(node)
    if isinstance(node, str): return node if is_numeric(node) else str(node)
    if isinstance(node, (int, float)): return node
    if isinstance(node, bool): return node
    if isinstance(node, complex): return str(node)
    if isinstance(node, list): return '[list]' if level >= self.depth else bracket(de_ve(self, node[:], level))
    if isinstance(node, dict): return '[dict]' if level >= self.depth else brace(de_en(self, list(node.items()), level))
    if isinstance(node, tuple): return '[tuple]' if level >= self.depth else parenth(de_ve(self, list(node[:]), level))
    if inspect.isfunction(node): return str(node)
    if inspect.isclass(type(node)): return de_ob(self, node, level)
    return str(node)


def de_ve(self, vector, lv):
    mutate(vector, lambda v: str(deco_node(self, v, lv + 1)))
    if self.presets: fluo_vector(vector, self.presets, mutate=True)
    return render_vector(self, vector, lv)


def de_en(self, entries, lv):
    key_wd = mutate_key_pad(entries)
    mutate_values(entries, lambda v: str(deco_node(self, v, lv + 1, key_wd)))
    if self.presets: fluo_entries(entries, self.presets, mutate=True)
    return render_entries(self, entries, lv)


def de_ob(self, ob, lv):
    abstract = ob.__dict__ \
        if hasattr(ob, '__dict__') \
        else {s: getattr(ob, s) for s in ob.__slots__} \
        if hasattr(ob, '__slots__') \
        else str(ob)
    return f'[{type(ob).__name__}] {deco_node(self, abstract, lv)}'



from typing import Callable, List, Tuple

from ject import oneself
from palett import Preset, fluo_entries
from palett.presets import FRESH, PLANET
from texting import COLF, COSP, ELLIP, LF, liner, to_br
from texting.enum.brackets import BRK, PAR
from veho.entries import zipper
from veho.vector import mapper
from veho.vector.length import length

from pyspare.margin import EntriesMargin
from pyspare.padder.pad_entries import pad_entries


def deco_entries(
        entries: list,
        key_read: Callable = None,
        read: Callable = None,
        head: int = None,
        tail: int = None,
        presets: Tuple[Preset] = (FRESH, PLANET),
        effects: List[str] = None,
        delim: str = COLF,
        bracket: int = BRK,
        inner_bracket: int = PAR,
        ansi: bool = False,
        dash: str = COSP
):
    size = length(entries)
    if not size: return str(entries)
    vn = EntriesMargin.build(entries, head, tail)
    raw, text = vn.to_list(ELLIP), vn.stringify(key_read, read).to_list(ELLIP)
    dye = fluo_entries(raw, presets, effects, colorant=True, mutate=True) if presets else None
    entries = pad_entries(text, raw, dye, ansi=presets or ansi) \
        if delim.find(LF) >= 0 \
        else zipper(text, dye, lambda tx, dy: dy(tx)) if presets else text
    brk = to_br(inner_bracket) or oneself
    lines = mapper(entries, lambda entry: brk(entry[0] + dash + entry[1].rstrip()))
    return liner(lines, delim=delim, bracket=bracket)



from ject import oneself
from texting import ELLIP

from pyspare.margin import VectorMargin


def margin_set(vec, head, tail, read, hr=ELLIP):
    vn = VectorMargin.build(vec, head, tail)
    return (
        vn.map(oneself).to_list(hr),
        vn.stringify(read).to_list(hr)
    )



from ject import oneself
from palett import fluo_matrix
from palett.presets import FRESH, PLANET
from texting import COLF, COSP, ELLIP, bracket as bracket_fn, liner
from texting.enum.brackets import BRK
from veho.enum.matrix_directions import ROWWISE
from veho.matrix import size
from veho.vector import mapper

from pyspare.margin import MatrixMargin
from pyspare.padder.pad_matrix import pad_matrix


def deco_matrix(matrix,
                discrete=False,
                delim=COSP,
                bracket=BRK,
                read=None,
                presets=(FRESH, PLANET),
                direct=ROWWISE,
                top=0,
                bottom=0,
                left=0,
                right=0,
                ansi=False,
                level=0,
                hr=ELLIP):
    if not matrix: return str(matrix)
    height, width = size(matrix)
    if not height or not width: return liner([], delim, level, bracket, discrete)
    vn = MatrixMargin.build(matrix, top, bottom, left, right, height, width)
    raw, text = vn.map(oneself).to_matrix(hr), vn.stringify(read).to_matrix(hr)
    dye = fluo_matrix(raw, direct, presets, colorant=True) if presets else None
    rows = pad_matrix(text, raw, dye, ansi)
    lines = mapper(rows, lambda line: bracket_fn(delim.join(line))) \
        if bracket \
        else mapper(rows, lambda line: delim.join(line))
    return liner(lines, delim=COLF, bracket=bracket, level=level, discrete=discrete)



from typing import Callable, List, Tuple

from ject import oneself
from palett import Preset, fluo_vector
from palett.presets import FRESH, PLANET
from texting import ELLIP
from veho.vector import mutazip
from veho.vector.length import length

from texting.enum.brackets import BRK
from texting import liner
from pyspare.margin import VectorMargin


def deco_vector(
        vec: list,
        read: Callable = None,
        head: int = None,
        tail: int = None,
        presets: Tuple[Preset] = (FRESH, PLANET),
        effects: List[str] = None,
        delim=',\n',
        bracket=BRK,
        discrete=False
):
    size = length(vec)
    if not size: return str(vec)
    vn = VectorMargin.build(vec, head, tail)
    raw, text = vn.map(oneself).to_list(ELLIP), vn.stringify(read).to_list(ELLIP)
    if presets:
        dyes = fluo_vector(raw, presets, effects, colorant=True, mutate=True)
        text = mutazip(text, dyes, lambda tx, dy: dy(tx))
    return liner(text, delim=delim, bracket=bracket, discrete=discrete)



from dataclasses import dataclass

from veho.vector import margin_mapper, margin_mutate

from pyspare.margin.utils import marginal
from pyspare.margin.vector_margin.sizing import sizing


@dataclass
class VectorMargin:
    vec: list
    head: int
    tail: int

    @staticmethod
    def build(vec: list, head: int, tail: int) -> 'VectorMargin':
        h, t = sizing(vec, head, tail)
        return VectorMargin(vec, h, t)

    def map(self, func, mutate=False) -> 'VectorMargin':
        boot, mapper = (self.reboot, margin_mutate) if mutate else (self.clone, margin_mapper)
        return boot(mapper(self.vec, func, self.head, self.tail))

    def stringify(self, func=None, mutate=False) -> 'VectorMargin':
        fn = (lambda x: str(func(x))) if func else str
        return self.map(fn, mutate)

    def to_list(self, ellip=None):
        return marginal(self.vec, self.head, self.tail, ellip)
        # return (((ve[:h] + [ellip] + ve[-t:]) if ellip else (ve[:h] + ve[-t:]))
        #         if t
        #         else ve[:h]) if h else ((([ellip] + ve[-t:]) if ellip else ve[-t:])
        #                                 if t
        #                                 else [])

    def reboot(self, vec):
        if vec: self.vec = vec
        return self

    def clone(self, vec):
        return VectorMargin(vec, self.head, self.tail)



from texting import CO, COSP, RTSP, SP, lange
from veho.vector import mutate

from pyspare.deco.deco_node import mutate_key_pad
from texting import join_lines
from pyspare.padder.pad_string import to_lpad

lpad = to_lpad(fill=SP, ansi=True)


def render_entries(self, entries, lv):
    vert = 0 if self.vert is None else self.vert
    width = 0 if self.width is None else self.width
    unit = 0 if self.unit is None else self.unit
    rows = mutate(entries, lambda kv: lpad(kv[0], pad) + RTSP + kv[1]) \
        if (lv < vert or any(lange(v) > unit for _, v in entries) or not width) and (pad := mutate_key_pad(entries)) \
        else wrap_entries(entries, width)
    return join_lines(rows, CO, lv) \
        if len(rows) > 1 \
        else COSP.join(rows)


def wrap_entries(entries, width):
    lines = []
    row, wd, kvp, sp = None, 0, None, len(COSP)
    for k, v in entries:
        wd += lange(kvp := k + RTSP + v) + sp
        if row and wd > width:
            lines.append(COSP.join(row))
            row = None
        if not row: row, wd = [], 0
        row.append(kvp)
    if row and len(row): lines.append(COSP.join(row))
    return lines



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyspare',
 'pyspare.deco',
 'pyspare.deco.deco_entries',
 'pyspare.deco.deco_json',
 'pyspare.deco.deco_matrix',
 'pyspare.deco.deco_node',
 'pyspare.deco.deco_node.helpers',
 'pyspare.deco.deco_node.preset',
 'pyspare.deco.deco_node.render',
 'pyspare.deco.deco_str',
 'pyspare.deco.deco_vector',
 'pyspare.logger',
 'pyspare.margin',
 'pyspare.margin.entries_margin',
 'pyspare.margin.matrix_margin',
 'pyspare.margin.utils',
 'pyspare.margin.vector_margin',
 'pyspare.padder',
 'pyspare.padder.pad_entries',
 'pyspare.padder.pad_keyed_column',
 'pyspare.padder.pad_matrix',
 'pyspare.padder.pad_string',
 'pyspare.padder.pad_table',
 'pyspare.padder.pad_vector']

package_data = \
{'': ['*']}

install_requires = \
['aryth>=0.0.6',
 'intype>=0.0.2',
 'ject>=0.0.1',
 'palett>=0.0.4',
 'texting>=0.0.8',
 'veho>=0.0.4']

setup_kwargs = {
    'name': 'pyspare',
    'version': '0.0.1',
    'description': 'pretty print',
    'long_description': '## pyspare\n##### pretty print\n\n### Usage\n```python\nfrom pyspare import deco\nvect = [1,2,3]\nprint(deco(vect))\n```',
    'author': 'Hoyeung Wong',
    'author_email': 'hoyeungw@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pydget/pyspare',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from veho.vector import length


def sizing(vec, head, tail):
    size = length(vec)
    if not size:
        return 0, 0
    elif (not head and not tail) or (head >= size):
        return size, 0
    else:
        return head, tail



from dataclasses import dataclass

from veho.entries.margin import margin_mapper, margin_mutate

from pyspare.margin.vector_margin import VectorMargin, sizing


@dataclass
class EntriesMargin(VectorMargin):
    # def __init__(self):
    #     VectorMargin.__init__(self,)

    @staticmethod
    def build(vec: list, head: int, tail: int) -> 'EntriesMargin':
        h, t = sizing(vec, head, tail)
        return EntriesMargin(vec, h, t)

    def map(self, kfn, vfn=None, mutate=False) -> 'EntriesMargin':
        boot, mapper = (self.reboot, margin_mutate) if mutate else (self.clone, margin_mapper)
        return boot(mapper(self.vec, kfn, vfn, self.head, self.tail))

    def stringify(self, kfn, vfn=None, mutate=False) -> 'EntriesMargin':
        kf = (lambda x: str(kfn(x))) if kfn else str
        vf = (lambda x: str(vfn(x))) if vfn else kf
        return self.map(kf, vf, mutate)



from texting import CO, COSP, lange

from texting import join_lines


def render_vector(self, vector, lv):
    vert = 0 if self.vert is None else self.vert
    width = 0 if self.width is None else self.width
    unit = 0 if self.unit is None else self.unit
    rows = vector \
        if lv < vert or any(lange(x) > unit for x in vector) or not width \
        else wrap_vector(vector, width)
    return join_lines(rows, CO, lv) \
        if len(rows) > 1 \
        else COSP.join(vector)


def wrap_vector(vector, width):
    lines = []
    row, size, sp = None, 0, len(COSP)
    for item in vector:
        size += lange(item) + sp
        if row and size > width:
            lines.append(COSP.join(row))
            row = None
        if not row: row, size = [], 0
        row.append(item)
    return lines



import re
from functools import partial

from palett import fluo_vector
from palett.presets import ATLAS, SUBTLE
from texting import LF, TB, fold, has_ansi, ripper
from texting.enum.regexes import LITERAL

splitter = partial(ripper, re.compile(LITERAL))


def deco_str(
        text,
        width=80,
        indent=0,
        first_line_indent=0,
        presets=(ATLAS, SUBTLE),
        effects=None,
        vectify=splitter,
        joiner=None
):
    if not (size := len(text)): return ''
    if has_ansi(text): return text
    if width and size > width: text = fold(
        text,
        width=width,
        delim=LF + TB * (indent if indent else 0),
        first_line_indent=first_line_indent
    )
    if presets: text = fluo_string(text, presets, effects, vectify, joiner)
    return text


def fluo_string(text, presets, effects, vectify, joiner):
    words = vectify(text)
    fluo_vector(words, presets, effects, mutate=True)
    return joiner(words) if joiner else ''.join(words)
